/*
 *  TOPPERS Software Toyohashi Open Platform for Embedded Real-Time
 *      Systems
 * 
 *  Copyright (C) 2019 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: test_spinlock2.c 264 2019-01-09 06:45:04Z ertl-hiro $
 */

/* 
 *		スピンロック機能のテスト(2)
 *
 * 【テストの目的】
 *
 *	スピンロックを取得したまま処理単位からリターンした場合の振る舞いな
 *	どをテストする．
 *
 * 【テスト項目】
 *
 *	(A) タスクが，取得したスピンロックを返却せずに終了した場合，カーネ
 *		ルによってスピンロックが返却されること［NGKI5167］
 *		(A-1) ext_tskによる終了
 *	(B) 次の処理単位が，取得したスピンロックを返却せずに終了した場合，
 *		カーネルによってスピンロックが返却されること［NGKI5167］
 *		(B-1) カーネル管理の割込みハンドラ
 *			  ※最後の割込みサービスルーチンからのリターンでテストする．
 *		(B-2) 割込みサービスルーチン
 *		(B-3) 周期ハンドラ
 *		(B-4) アラームハンドラ
 *		(B-5) カーネル管理のCPU例外ハンドラ
 *	(C) 拡張サービスコールが，取得したスピンロックを返却せずに終了した
 *		場合，スピンロックは返却されないこと［NGKI2123］
 *	(D) スピンロックを取得している場合，unl_cpuはCPUロックを解除しない
 *		こと［NGKI2739］
 *
 * 【使用リソース】
 *
 *	CLS_PRC1: プロセッサ1のみで実行
 *	CLS_PRC2: プロセッサ2のみで実行
 *
 *	TASK1: CLS_PRC1，中優先度タスク，メインタスク，最初から起動
 *	TASK2: CLS_PRC1，高優先度タスク
 *	SPN1: CLS_PRC1，スピンロック
 *	CYC1: CLS_PRC1，周期ハンドラ
 *	ALM1: CLS_PRC1，アラームハンドラ
 *  ISR1: CLS_PRC1，割込みサービスルーチン
 *  ISR2: CLS_PRC1，割込みサービスルーチン
 *	CPUEXC1: CPU例外ハンドラ
 *	EXTSVC1: 拡張サービスコール
 *
 * 【テストシーケンス】
 *
 *	== START_PRC1 ==					｜	== START_PRC2 ==
 *	== TASK1（中）==					｜
 *	1:	act_tsk(TASK2)					｜
 *	== TASK2-1（高，1回目）==			｜
 *	2:	loc_spn(SPN1)					｜
 *		ext_tsk()			... (A-1)	｜
 *	== TASK1（続き）==					｜
 *	3:	DO(intno1_clear())				｜
 *		try_spn(SPN1)					｜
 *		unl_spn(SPN1)					｜
 *		ras_int(INTNO1)					｜
 *	== ISR1 ==							｜
 *	4:	DO(intno1_clear())				｜
 *		try_spn(SPN1)					｜
 *		RETURN				... (B-2)	｜
 *	== ISR2 ==							｜
 *	5:	try_spn(SPN1)					｜
 *		RETURN				... (B-1)	｜
 *	== TASK1（続き）==					｜
 *	6:	sta_cyc(CYC1)					｜
 *		slp_tsk()						｜
 *	== CYC1 ==							｜
 *	7:	wup_tsk(TASK1)					｜
 *		try_spn(SPN1)					｜
 *		RETURN				... (B-3)	｜
 *	== TASK1（続き）==					｜
 *	8:	try_spn(SPN1)					｜
 *		unl_spn(SPN1)					｜
 *		stp_cyc(CYC1)					｜
 *		sta_alm(ALM1, TEST_TIME_CP)		｜
 *		slp_tsk()						｜
 *	== ALM1 ==							｜
 *	9:	wup_tsk(TASK1)					｜
 *		try_spn(SPN1)					｜
 *		RETURN				... (B-4)	｜
 *	== TASK1（続き）==					｜
 *	10:	try_spn(SPN1)					｜
 *		unl_spn(SPN1)					｜
 *		cal_svc(TFN_EXTSVC1, 0, 0, 0, 0, 0)｜
 *	== EXTSVC1 ==						｜
 *	11:	try_spn(SPN1)					｜
 *		RETURN(E_OK)			... (C)	｜
 *	== TASK1（続き）==					｜
 *	12:	unl_cpu()				... (D)	｜
 *		assert(sns_loc() == true)		｜
 *		unl_spn(SPN1)					｜
 *		DO(RAISE_CPU_EXCEPTION)			｜
 *	== CPUEXC1 ==						｜
 *	13:	act_tsk(TASK2)					｜
 *		try_spn(SPN1)					｜
 *  	RETURN				... (B-5)	｜
 *	== TASK2-2（高，2回目）==			｜
 *	14:	try_spn(SPN1)					｜
 *		unl_spn(SPN1)					｜
 *	15:	END								｜
 */

#include <kernel.h>
#include <t_syslog.h>
#include "syssvc/test_svc.h"
#include "kernel_cfg.h"
#include "test_common.h"

/* DO NOT DELETE THIS LINE -- gentest depends on it. */

void
alarm1_handler(intptr_t exinf)
{
	ER_UINT	ercd;

	check_point_PRC1(9);
	ercd = wup_tsk(TASK1);
	check_ercd(ercd, E_OK);

	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return;

	check_assert(false);
}

void
cpuexc1_handler(void *p_excinf)
{
	ER_UINT	ercd;

	check_point_PRC1(13);
	ercd = act_tsk(TASK2);
	check_ercd(ercd, E_OK);

	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return;

	check_assert(false);
}

void
cyclic1_handler(intptr_t exinf)
{
	ER_UINT	ercd;

	check_point_PRC1(7);
	ercd = wup_tsk(TASK1);
	check_ercd(ercd, E_OK);

	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return;

	check_assert(false);
}

ER_UINT
extsvc1_routine(intptr_t par1, intptr_t par2, intptr_t par3,
								intptr_t par4, intptr_t par5, ID cdmid)
{
	ER_UINT	ercd;

	check_point_PRC1(11);
	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return(E_OK);

	check_assert(false);
	return(E_SYS);
}

void
isr1(intptr_t exinf)
{
	ER_UINT	ercd;

	check_point_PRC1(4);
	intno1_clear();

	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return;

	check_assert(false);
}

void
isr2(intptr_t exinf)
{
	ER_UINT	ercd;

	check_point_PRC1(5);
	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	return;

	check_assert(false);
}

void
task1(intptr_t exinf)
{
	ER_UINT	ercd;

	test_start(__FILE__);

	check_point_PRC1(1);
	ercd = act_tsk(TASK2);
	check_ercd(ercd, E_OK);

	check_point_PRC1(3);
	intno1_clear();

	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = unl_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = ras_int(INTNO1);
	check_ercd(ercd, E_OK);

	check_point_PRC1(6);
	ercd = sta_cyc(CYC1);
	check_ercd(ercd, E_OK);

	ercd = slp_tsk();
	check_ercd(ercd, E_OK);

	check_point_PRC1(8);
	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = unl_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = stp_cyc(CYC1);
	check_ercd(ercd, E_OK);

	ercd = sta_alm(ALM1, TEST_TIME_CP);
	check_ercd(ercd, E_OK);

	ercd = slp_tsk();
	check_ercd(ercd, E_OK);

	check_point_PRC1(10);
	ercd = try_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = unl_spn(SPN1);
	check_ercd(ercd, E_OK);

	ercd = cal_svc(TFN_EXTSVC1, 0, 0, 0, 0, 0);
	check_ercd(ercd, E_OK);

	check_point_PRC1(12);
	ercd = unl_cpu();
	check_ercd(ercd, E_OK);

	check_assert(sns_loc() == true);

	ercd = unl_spn(SPN1);
	check_ercd(ercd, E_OK);

	RAISE_CPU_EXCEPTION;

	check_assert(false);
}

static uint_t	task2_count = 0;

void
task2(intptr_t exinf)
{
	ER_UINT	ercd;

	switch (++task2_count) {
	case 1:
		check_point_PRC1(2);
		ercd = loc_spn(SPN1);
		check_ercd(ercd, E_OK);

		ercd = ext_tsk();
		check_ercd(ercd, E_OK);

		check_assert(false);

	case 2:
		check_point_PRC1(14);
		ercd = try_spn(SPN1);
		check_ercd(ercd, E_OK);

		ercd = unl_spn(SPN1);
		check_ercd(ercd, E_OK);

		check_finish_PRC1(15);
		check_assert(false);

	default:
		check_assert(false);
	}
	check_assert(false);
}
